use crate::config::Config;
use crate::patterns::{Patterns, PixelColor, Pattern};
use crate::Strip;
use bluer::Address;
use serde::{Deserialize, Serialize};
use std::collections::HashMap;
use std::future::Future;
use std::str::FromStr;
use std::time::Duration;
use tokio::sync::watch::{Receiver, Sender};

pub(crate) mod period {
    use std::time::Duration;
    use serde::{de::Error as _, Deserialize, Deserializer};


    /// Custom deserializer for period from millisecond value
    ///
    /// # Arguments
    ///
    /// * `deserializer`:
    ///
    /// returns: Result<Duration, <D as Deserializer>::Error>
    ///
    pub fn deserialize<'de, D>(deserializer: D) -> Result<Duration, D::Error>
        where
            D: Deserializer<'de>,
    {
        String::deserialize(deserializer)?
            .parse::<u64>().and_then(|d| Ok(Duration::from_millis(d)))
            .map_err(|e| D::Error::custom(format!("{}", e)))
    }
}

#[derive(Serialize, Deserialize, Debug, Eq, Hash, PartialEq, Copy, Clone)]
pub struct Context<const N: usize> {
    pub(crate) pattern: Patterns<N>,
    #[serde(deserialize_with  = "period::deserialize")]
    pub period: Duration,
}

pub(crate) type DeviceSequence<const N: usize> = Vec<Context<N>>;



pub async fn runner_loop<const N: usize>(runner: DeviceSequence<N>, tx: Sender<Strip<N>>) {
    let mut background = Strip::<N>::default();
    for mut context in runner {
        let mut strip = context.pattern.init().unwrap_or(background);
        let delay = context.period;
        while tx.send(strip).is_ok() {
            if let Some(value) = context.pattern.next() {
                strip = apply_mask(background, value);
            } else {
                break;
            }
            tokio::time::sleep(delay).await;
        }
        background = strip;
        println!("{:?}", background);
    }
}

pub fn apply_mask<const N: usize>(previous_strip: Strip<N>, strip: Strip<{ N }>) -> Strip<N> {
    let mut new_strip = Strip::default();
    for i in 0..N {
        if let Some(color) = strip[i] {
            new_strip[i] = Some(color);
        } else {
            new_strip[i] = previous_strip[i];
        }
    }
    new_strip
}

#[derive(Default)]
pub struct Spectacle<const N: usize> {
    runners: HashMap<DeviceSequence<N>, (Sender<Strip<N>>, Receiver<Strip<N>>)>,
    channels: HashMap<Address, Receiver<Strip<N>>>,
    default_runner: Option<(DeviceSequence<N>, Sender<Strip<N>>)>,
    default_receiver: Option<Receiver<Strip<N>>>,
}

impl<const N: usize> Spectacle<N> {
    pub fn add(&mut self, mac_address: Address, runner: &DeviceSequence<N>) {
        if let Some((_, rx)) = self.runners.get(runner as &DeviceSequence<N>) {
            self.channels.insert(mac_address, rx.clone());
        } else {
            let (tx, rx) = tokio::sync::watch::channel(Strip::<N>::new(PixelColor {
                red: 0,
                green: 255,
                blue: 0,
            }));
            self.runners.insert(runner.clone(), (tx, rx.clone()));
            self.channels.insert(mac_address, rx);
        }
    }

    pub fn add_default(&mut self, runner: DeviceSequence<N>) {
        let (tx, rx) = tokio::sync::watch::channel(Strip::<N>::new(PixelColor {
            red: 0,
            green: 0,
            blue: 255,
        }));
        // self.tasks.push(runner.runner_loop(tx));
        self.default_runner = Some((runner, tx));
        self.default_receiver = Some(rx);
    }

    // pub fn nbr_tasks(&self) -> usize {
    //     let mut nbr = 0;
    //     if self.default_receiver.is_some() {
    //         nbr += 1;
    //     }
    //     nbr += self.runners.len();
    //     nbr
    // }

    pub fn get_channel(&self, mac_address: &Address) -> Option<Receiver<Strip<N>>> {
        let channel = self.channels.get(mac_address);
        match channel {
            None => {
                let d = &self.default_receiver;
                match d {
                    None => None,
                    Some(runner) => {
                        println!("default rx");
                        Some(runner.clone())
                    }
                }
            }
            Some(channel) => {
                println!("rx found ");
                Some(channel.clone())
            }
        }
    }

    pub fn run(&mut self) -> Vec<impl Future<Output = ()>> {
        let mut joinhandles = vec![];

        if self.default_runner.is_some() {
            let d = std::mem::take(&mut self.default_runner).unwrap();
            joinhandles.push(runner_loop(d.0.clone(), d.1));
            println!("add default runner loop");
        }

        for (runner, (tx, _)) in self.runners.drain().take(1) {
            joinhandles.push(runner_loop(runner.clone(), tx));
            println!("add custom runner loop");
        }
        joinhandles
    }

}

impl<const N: usize> From<Config<N>> for Spectacle<N> {
    fn from(config: Config<N>) -> Self {
        let mut spectacle: Spectacle<N> = Spectacle::default();
        for device in config {
            if let Some(mac_addresses) = device.mac_addresses {
                for mac_address in mac_addresses {
                    let mac_address = Address::from_str(&*mac_address);
                    match mac_address {
                        Ok(address) => {
                            spectacle.add(address, &device.sequence);
                        }
                        Err(_) => {
                            println!("error reading mac address");
                        }
                    }
                }
            } else {
                spectacle.add_default(device.sequence);
            }
        }
        spectacle
    }
}


#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn apply_mask() {
        let pixel_entry = PixelColor {
            red: 10,
            green: 10,
            blue: 10
        };

        let pixel_mask = PixelColor {
            red: 20,
            green: 20,
            blue: 20
        };

        let mut entry = Strip::<5>::new(pixel_entry) ;
        entry[0] = None;
        entry[1] = None;

        let mut mask = Strip::<5>::new(pixel_mask) ;
        mask[0] = None;

        let mut expected = entry.clone();
        expected[1] = Some(pixel_mask);
        assert_eq!(expected, super::apply_mask(mask, entry));
    }
}